"""Coordinate wrapper."""

from __future__ import annotations

import typing as ty
from functools import lru_cache
from pathlib import Path

import numpy as np
from koyo.timer import MeasureTimer
from koyo.typing import PathLike

from image2image_io.config import CONFIG
from image2image_io.readers._base_reader import BaseReader
from image2image_io.utils.lazy import LazyImageWrapper
from image2image_io.utils.utilities import format_mz

if ty.TYPE_CHECKING:
    from imzy._readers._base import BaseReader as BaseImzyReader


def set_dimensions(reader: CoordinateImageReader | LazyCoordinateImageReader) -> None:
    """Set dimension information."""
    x, y = reader.x, reader.y
    reader.xmin, reader.xmax = np.min(x), np.max(x)
    reader.ymin, reader.ymax = np.min(y), np.max(y)
    reader._image_shape = (reader.ymax - reader.ymin + 1, reader.xmax - reader.xmin + 1)


def get_image(array_or_reader: np.ndarray | BaseImzyReader) -> np.ndarray:
    """Return image for the array/image."""
    if not isinstance(array_or_reader, np.ndarray):
        array_or_reader = array_or_reader.reshape(array_or_reader.get_tic())
    return array_or_reader.astype(np.float32)  # type: ignore[no-any-return]


class CoordinateImagerMixin:
    """Mixin class to reduce amount of duplicate code."""

    xmin: int
    xmax: int
    ymin: int
    ymax: int
    x: np.ndarray
    y: np.ndarray
    is_fixed: bool = False
    image_shape: tuple[int, int]

    def get_random_image(self) -> np.ndarray:
        """Return random ion image."""
        array = np.full(self.image_shape, np.nan)
        array[self.y - self.ymin, self.x - self.xmin] = np.random.randint(128, 255, size=len(self.x)) / 255
        return array


class CoordinateImageReader(BaseReader, CoordinateImagerMixin):  # type: ignore[misc]
    """Reader for data that has defined coordinates."""

    def __init__(
        self,
        path: PathLike,
        x: np.ndarray,
        y: np.ndarray,
        key: str | None = None,
        resolution: float = 1.0,
        array_or_reader: np.ndarray | BaseImzyReader | None = None,
        data: dict[str, np.ndarray] | None = None,
        auto_pyramid: bool | None = None,
        reader_kws: dict | None = None,
    ):
        super().__init__(path, key, auto_pyramid=auto_pyramid, reader_kws=reader_kws)
        self.x = x
        self.y = y
        self.resolution = resolution
        self.reader = None if isinstance(array_or_reader, np.ndarray) else array_or_reader
        self.allow_extraction = self.reader is not None
        self._is_rgb = False
        self._array_dtype = np.float32  # type: ignore[assignment]

        self.data = data or {}
        if self.name not in self.data:
            name = "tic" if self.reader is not None else self.name
            self.data[name] = get_image(array_or_reader)
            self.get_image.cache_clear()  # clear cache
        set_dimensions(self)

    @property
    def _channel_names(self) -> list[str]:
        return list(self.data.keys())

    @property
    def channel_names(self) -> list[str]:
        """List of channel names."""
        return list(self.data.keys())

    @property
    def pyramid(self) -> list:
        """Pyramid."""
        return self.get_dask_pyr()

    def extract(self, mzs: np.ndarray, ppm: float = 10.0) -> tuple[Path, list[str]]:
        """Extract ion images."""
        if self.reader is None:
            raise ValueError("Cannot extract ion images from a numpy array.")
        mzs = np.atleast_1d(mzs)
        CONFIG.trace(f"Extracting {len(mzs)} ion images from {self.path.name} ({self.name})")

        with MeasureTimer() as timer:
            images = self.reader.get_ion_images(mzs, ppm=ppm)
        CONFIG.trace(f"Extracted {len(mzs)} ion images in {timer}")

        labels = []
        for i, mz in enumerate(mzs):
            label = format_mz(mz)
            self.data[label] = images[i]
            labels.append(f"{label} | {self.name}")
        self.get_image.cache_clear()  # clear cache
        return self.path, labels

    def get_dask_pyr(self) -> list[np.ndarray]:
        """Get dask representation of the pyramid."""
        return [self.get_image()]

    def get_channel_axis_and_n_channels(self, shape: tuple[int, ...] | None = None) -> tuple[int | None, int]:
        """Return channel axis and number of channels."""
        if shape is not None:
            return super().get_channel_axis_and_n_channels(shape)
        if len(self.data) == 1:
            return None, 1
        return 2, len(self.data)

    @lru_cache(maxsize=1)
    def get_image(self):
        """Return image as a stack."""
        if len(self.data) == 1:
            key = next(iter(self.data))
            return self.data[key]
        return np.dstack([self.data[key] for key in self.data])

    def get_channel(self, index: int, pyramid: int = 0, split_rgb: bool | None = None) -> np.ndarray:
        """Return channel."""
        image = self.get_channel_pyramid(index)[pyramid]
        return image

    def get_channel_pyramid(self, index: int) -> list[np.ndarray]:
        """Return channel pyramid."""
        name = self.channel_names[index]
        return [self.preprocessor("", self.data[name])]


class LazyCoordinateImageReader(BaseReader, CoordinateImagerMixin):  # type: ignore[misc]
    """Lazy coordinate image reader."""

    lazy = True

    def __init__(
        self,
        path: PathLike,
        x: np.ndarray,
        y: np.ndarray,
        lazy_wrapper: LazyImageWrapper,
        key: str | None = None,
        resolution: float = 1.0,
        channel_names: list[str] | None = None,
        auto_pyramid: bool | None = None,
    ):
        super().__init__(path, key, auto_pyramid=auto_pyramid)
        self.x = x
        self.y = y
        self.resolution = resolution
        self._is_rgb = False
        self._array_dtype = lazy_wrapper.dtype
        self._array_shape = (y.max(), x.max(), lazy_wrapper.shape[-1])
        self.lazy_wrapper = lazy_wrapper
        if channel_names:
            self._channel_names = channel_names
        set_dimensions(self)

    def get_dask_pyr(self) -> list[np.ndarray]:
        """Get dask representation of the pyramid."""
        return [self.get_image()]

    def get_image(self) -> np.ndarray:
        """Return image as a stack."""
        return self.lazy_wrapper.get_image()

    def get_channel_axis_and_n_channels(self, shape: tuple[int, ...] | None = None) -> tuple[int | None, int]:
        """Return channel axis and number of channels."""
        if shape:
            return super().get_channel_axis_and_n_channels(shape)
        return 2, len(self.channel_names)

    def get_channel_pyramid(self, index: int) -> list[np.ndarray]:
        """Return channel pyramid."""
        return [self.preprocessor("", self.lazy_wrapper[index])]



"""Tiff file wrapper.

Copied from:
https://github.com/NHPatterson/napari-imsmicrolink/blob/master/src/napari_imsmicrolink/data/tifffile_reader.py
"""

from __future__ import annotations

import warnings
from pathlib import Path

from koyo.typing import PathLike
from loguru import logger
from ome_types import from_xml
from tifffile import TiffFile

from image2image_io.config import CONFIG
from image2image_io.readers._base_reader import BaseReader
from image2image_io.readers.tiff_utils import (
    ometiff_ch_names,
    ometiff_xy_pixel_sizes,
    qptiff_channel_names,
    svs_xy_pixel_sizes,
    tifftag_xy_pixel_sizes,
)
from image2image_io.readers.utilities import get_tifffile_info, tifffile_to_dask
from image2image_io.utils.utilities import guess_rgb

logger = logger.bind(src="Tiff")


class TiffImageReader(BaseReader):
    """TIFF image wrapper."""

    fh: TiffFile
    reader = "tifffile"

    def __init__(
        self,
        path: PathLike,
        key: str | None = None,
        init_pyramid: bool | None = None,
        auto_pyramid: bool | None = None,
    ):
        super().__init__(path, key, auto_pyramid=auto_pyramid)
        self.fh = TiffFile(self.path)

        self.shape, self._array_dtype, self.largest_series = self._get_image_info()
        self._image_shape = self.shape[0:2] if self.is_rgb else (self.shape[1::] if len(self.shape) > 2 else self.shape)

        self.resolution = self._get_im_res()
        self._channel_names = self._get_channel_names()
        self._channel_colors = None
        CONFIG.trace(
            f"{path}: RGB={self.is_rgb}; dims={self.shape}; px={self.resolution:.3f}; n_ch={len(self._channel_names)}"
        )
        init_pyramid = init_pyramid if init_pyramid is not None else CONFIG.init_pyramid
        if init_pyramid:
            self._pyramid = self.pyramid

    @property
    def n_channels(self) -> int:
        """Return number of channels."""
        if self.is_rgb:
            return self.shape[2]
        _, n_channels = self.get_channel_axis_and_n_channels()
        return n_channels

    def get_dask_pyr(self) -> list:
        """Get instance of Dask pyramid."""
        d_pyr = tifffile_to_dask(self.path, self.largest_series)
        channel_axis, _ = self.get_channel_axis_and_n_channels(shape=d_pyr[0].shape)
        if self.is_rgb and guess_rgb(d_pyr[0].shape):
            d_pyr[0] = d_pyr[0].rechunk((2048, 2048, 1))
        elif len(d_pyr[0].shape) > 2:
            if channel_axis == 0:
                d_pyr[0] = d_pyr[0].rechunk((1, 2048, 2048))
            else:
                d_pyr[0] = d_pyr[0].rechunk((2048, 2048, 1))
        else:
            d_pyr[0] = d_pyr[0].rechunk((2048, 2048))
        return d_pyr

    def _get_im_res(self) -> float:
        if Path(self.path).suffix.lower() in [".scn", ".ndpi"]:
            return tifftag_xy_pixel_sizes(
                self.fh,
                self.largest_series,
                0,
            )[0]
        elif Path(self.path).suffix.lower() in [".svs"]:
            return svs_xy_pixel_sizes(
                self.fh,
                self.largest_series,
                0,
            )[0]
        elif self.fh.ome_metadata:
            return ometiff_xy_pixel_sizes(
                from_xml(self.fh.ome_metadata, parser="lxml"),
                self.largest_series,
            )[0]
        else:
            try:
                return tifftag_xy_pixel_sizes(
                    self.fh,
                    self.largest_series,
                    0,
                )[0]
            except KeyError:
                warnings.warn(
                    "Unable to parse pixel resolution information from file defaulting to 1",
                    stacklevel=2,
                )
                return 1.0

    def _get_channel_names(self) -> list[str]:
        channel_names = []
        if any(suffix in self.path.name for suffix in [".qptiff", ".qptiff.intermediate", ".qptiff.raw"]):
            channel_names = qptiff_channel_names(self.fh)
        if self.fh.ome_metadata and not channel_names:
            channel_names = ometiff_ch_names(from_xml(self.fh.ome_metadata, parser="lxml"), self.largest_series)
            if self.n_channels > len(channel_names):
                channel_names = []

        if not channel_names or len(channel_names) != self.n_channels:
            if self.is_rgb:
                channel_names = ["R", "G", "B"] if CONFIG.split_rgb else ["RGB"]
            else:
                channel_names = []
                for idx, _ch in enumerate(range(self.n_channels)):
                    channel_names.append(f"C{str(idx + 1).zfill(2)}")
        return channel_names

    def _get_image_info(self) -> tuple:
        if len(self.fh.series) > 1:
            warnings.warn(
                "The tiff contains multiple series, the largest series will be read by default",
                stacklevel=2,
            )

        array_shape, _, array_dtype, largest_series = get_tifffile_info(self.path)
        if not CONFIG.auto_pyramid:
            largest_series = 0
        return array_shape, array_dtype, largest_series

    def get_channel_axis_and_n_channels(self, shape: tuple | None = None) -> tuple[int | None, int]:
        """Return channel axis and number of channels."""
        if shape is None:
            shape = self.shape
        ndim = len(shape)
        # 2D images will be returned as they are
        if ndim == 3:
            if self.is_rgb:
                return 2, 3
            # elif np.argmin(shape) == 2:
            #     return 2, shape[2]
            else:
                return 0, shape[0]
        return None, 1



"""Convert images from one format to OME-TIFF format."""

from __future__ import annotations

import click
from koyo.click import arg_split_int, arg_split_str
from loguru import logger


@click.option(
    "-N",
    "--channel_names",
    type=click.STRING,
    default=None,
    help="Specify a list of channel names to be used by the task. Names should be separated by commas. "
    "e.g. 'DAPI,CD3,CD20'",
    callback=arg_split_str,
    show_default=True,
    required=False,
)
@click.option(
    "-I",
    "--channel_ids",
    type=click.STRING,
    default=None,
    help="Specify a list of channel indices to be used by the task. Channel IDs should be separated by comma."
    "e.g. '1,2,4'.",
    callback=arg_split_int,
    show_default=True,
    required=False,
)
@click.option(
    "-u",
    "--as_uint8",
    is_flag=True,
    help="Convert to uint8. If not specified, the original data type will be used.",
    show_default=True,
    required=False,
)
@click.option(
    "-t",
    "--tile_size",
    help="Tile size.",
    type=click.Choice(["256", "512", "1024", "2048"], case_sensitive=False),
    default="512",
    show_default=True,
    required=False,
)
@click.option(
    "-s",
    "--scene",
    "scene_index",
    type=click.INT,
    help="Specify the scene to be processed. If not specified, all scenes will be processed.",
    default=None,
    show_default=True,
    required=True,
)
@click.option(
    "-o",
    "--output_dir",
    help="Path to directory where OME-TIFF files will be saved.",
    default=".",  # cwd
    type=click.Path(exists=False, resolve_path=True, file_okay=False, dir_okay=True),
    show_default=True,
    required=True,
)
@click.option(
    "-i",
    "--input",
    "input_",
    help="Path to the CZI file that should be converted.",
    type=click.Path(exists=False, resolve_path=False, file_okay=True, dir_okay=False),
    show_default=True,
    required=True,
)
@click.command()
def czi2tiff(
    input_: str,
    output_dir: str,
    scene_index: int,
    tile_size: str,
    as_uint8: bool,
    channel_ids: list[int] | None,
    channel_names: list[str | None],
) -> None:
    """Convert CZI to OME-TIFF."""
    from image2image_io.writers import czi_to_ome_tiff

    metadata = None
    if channel_ids and channel_names:
        if len(channel_ids) != len(channel_names):
            raise ValueError("Number of channel IDs and channel names must be equal.")
        metadata = {scene_index: {"channel_ids": channel_ids, "channel_names": channel_names}}

    for key, scene_index, total, _ in czi_to_ome_tiff(
        input_, output_dir, as_uint8=as_uint8, tile_size=int(tile_size), metadata=metadata, scenes=[scene_index]
    ):
        logger.info(f"Converted {key} scene {scene_index}/{total}")



"""Writer tests."""

import tempfile

import numpy as np

from image2image_io.readers import ArrayImageReader
from image2image_io.writers import OmeTiffWriter

tmp = tempfile.TemporaryDirectory()
tmp_path = tmp.name

array = np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint8)
reader = ArrayImageReader(tmp_path, array)
assert reader.is_rgb, "Array should be rgb"
writer = OmeTiffWriter(reader)
path = writer.write("test", tmp_path)
assert path.exists(), "Path should exist"
assert path.is_file(), "Path should be a file"

tmp.cleanup()



"""GeoJSON reader for image2image."""

from __future__ import annotations

import typing as ty
from pathlib import Path

import numpy as np
import pandas as pd
from koyo.rand import temporary_seed
from koyo.typing import PathLike
from loguru import logger

from image2image_io.config import CONFIG
from image2image_io.readers._base_reader import BaseReader
from image2image_io.readers.utilities import get_column_name


def read_points(path: PathLike, return_df: bool = False) -> tuple[np.ndarray, np.ndarray, pd.DataFrame] | pd.DataFrame:
    """Read points from CSV/parquet file."""
    path = Path(path)
    if path.suffix == ".csv":
        df = pd.read_csv(path)
    elif path.suffix in [".txt", ".tsv"]:
        temp = pd.read_csv(path, delimiter="\t", nrows=1)
        sep = "\t" if len(temp.columns) > 1 else " "
        df = pd.read_csv(path, delimiter=sep)
    elif path.suffix == ".parquet":
        df = pd.read_parquet(path)
    else:
        raise ValueError(f"Invalid file extension: {path.suffix}")
    if return_df:
        return df
    return read_points_from_df(df)


def read_points_from_df(df: pd.DataFrame) -> tuple[np.ndarray, np.ndarray, pd.DataFrame]:
    """Read points from DataFrame."""
    x_key = get_column_name(df, ["x", "x_location", "x_centroid", "x:x"])
    y_key = get_column_name(df, ["y", "y_location", "y_centroid", "y:y"])
    if x_key not in df.columns or y_key not in df.columns:
        raise ValueError(f"Invalid columns: {df.columns}")
    x = df[x_key].values
    y = df[y_key].values
    return x, y, df.drop(columns=[x_key, y_key])


def get_channel_names_from_df(df: pd.DataFrame) -> tuple[str, list[str]]:
    """Return list of column names. This is very specific and might break under certain circumstances."""
    filter_col = get_column_name(df, ["feature_name", "name"])
    if not filter_col:
        return "", list(df.columns)
    return filter_col, list(df[filter_col].unique())


class PointsReader(BaseReader):
    """GeoJSON reader for image2image."""

    reader_type = "points"
    _channel_names: list[str]
    _is_rgb: bool = False
    _array_dtype = np.dtype("float32")

    def __init__(
        self,
        path: PathLike,
        key: str | None = None,
        auto_pyramid: bool | None = None,
    ):
        super().__init__(path, key=key, auto_pyramid=auto_pyramid)

        self.x, self.y, self.df = read_points(self.path)  # type: ignore[var-annotated]
        # self.filter_column, self._channel_names = get_channel_names_from_df(self.df)
        self._channel_names = list(self.df.columns)

    def to_points(self) -> tuple[str, dict[str, np.ndarray | str]]:
        """Convert to shapes that can be exported to Shapes layer."""
        x, y, df = self.parse_data()
        return "points", {"data": np.c_[y, x], "properties": df}

    def parse_data(self) -> tuple:
        """Parse data."""
        x, y, df = self.x, self.y, self.df
        n_points = len(x)
        if n_points > 1_000_000 and CONFIG.subsample and CONFIG.subsample_ratio < 1.0:
            n_subsample = int(CONFIG.subsample_ratio * n_points)
            logger.trace(f"Subsampling to {n_subsample:,} points.")
            with temporary_seed(CONFIG.subsample_random_seed, skip_if_negative_one=True):
                indices = np.random.choice(n_points, n_subsample, replace=False)
            x = x[indices]
            y = y[indices]
            df = df.iloc[indices]
        return x, y, df

    def to_points_kwargs(self, face_color: str, **kwargs: ty.Any) -> dict:
        """Return data so it's compatible with Shapes layer."""
        x, y, df = self.parse_data()
        n = len(x)

        kws = {
            "data": np.c_[y, x],
            "scale": self.scale,
            "affine": self.transform,
            "features": df,
            "face_color": face_color,
            "size": 5 if n < 50_000 else 1,
        }
        kws.update(kwargs)
        return kws

    def get_dask_pyr(self) -> list[ty.Any]:
        """Get dask representation of the pyramid."""
        raise NotImplementedError("Must implement method")

    def to_csv(self, path: PathLike) -> str:
        """Export data as CSV file."""
        raise NotImplementedError("Must implement method")

    @property
    def shape(self) -> tuple[int, ...]:
        """Return shape of data."""
        return (len(self.x),)



"""Test utilities."""

import pandas as pd
import pytest

from image2image_io.readers.utilities import check_df_columns


@pytest.mark.parametrize("columns", [["x", "y"], ["x_location", "y_location"], ["x", "y_location"]])
def test_if_points(columns):
    x_col, y_col = columns
    df = pd.DataFrame({x_col: [1, 2, 3], y_col: [4, 5, 6]})
    assert check_df_columns(df, [x_col, y_col]), "Expected columns to be valid."


@pytest.mark.parametrize(
    "columns",
    [
        ["x", "y", ("cell", "cell_id")],
        ["vertex_x", "vertex_y", ("cell", "cell_id")],
        ["x", "vertex_y", ("cell", "cell_id")],
    ],
)
def test_if_shapes(columns):
    x_col, y_col, either = columns
    df = pd.DataFrame({x_col: [1, 2, 3], y_col: [4, 5, 6], either[0]: ["a", "b", "c"]})
    assert check_df_columns(df, [x_col, y_col], [either]), "Expected columns to be valid."

    df = pd.DataFrame({x_col: [1, 2, 3], y_col: [4, 5, 6], either[1]: ["a", "b", "c"]})
    assert check_df_columns(df, [x_col, y_col], [either]), "Expected columns to be valid."



"""Enums."""

import typing as ty
from enum import Enum

import numpy as np

DEFAULT_TRANSFORM_NAME: str = "Identity matrix"

TIME_FORMAT = "%d/%m/%Y-%H:%M:%S:%f"

MaskOutputFmt = ty.Literal["hdf5", "binary", "geojson"]
WriterMode = ty.Literal["sitk", "ome-zarr", "ome-tiff", "ome-tiff-by-plane", "ome-tiff-by-tile"]
DisplayType = ty.Literal["points", "polygon", "path", "path or polygon"]


class ViewType(str, Enum):
    """View type."""

    RANDOM = "random"
    OVERLAY = "overlay"


SITK_TO_NP_DTYPE = {
    0: np.int8,
    1: np.uint8,
    2: np.int16,
    3: np.uint16,
    4: np.int32,
    5: np.uint32,
    6: np.int64,
    7: np.uint64,
    8: np.float32,
    9: np.float64,
    10: np.complex64,
    11: np.complex64,
    12: np.int8,
    13: np.uint8,
    14: np.int16,
    15: np.int16,
    16: np.int32,
    17: np.int32,
    18: np.int64,
    19: np.int64,
    20: np.float32,
    21: np.float64,
    22: np.uint8,
    23: np.uint16,
    24: np.uint32,
    25: np.uint64,
}



"""Writer tests."""

import numpy as np
import pytest

from image2image_io.models import MergeImages
from image2image_io.readers import ArrayImageReader, TiffImageReader
from image2image_io.writers import MergeOmeTiffWriter, OmeTiffWrapper, OmeTiffWriter


def test_write_rgb_array(tmp_path):
    array = np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint8)
    reader = ArrayImageReader(tmp_path, array)
    assert reader.is_rgb, "Array should be rgb"
    writer = OmeTiffWriter(reader)
    path = writer.write("test", tmp_path)
    assert path.exists(), "Path should exist"
    assert path.is_file(), "Path should be a file"


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_write_array_dtype(tmp_path, dtype):
    array = np.random.random((3, 1024, 1024)).astype(dtype)
    reader = ArrayImageReader(tmp_path, array)
    assert reader.n_channels == 3, "Array should have 3 channels"
    assert not reader.is_rgb, "Array should not be rgb"
    writer = OmeTiffWriter(reader)
    path = writer.write("test", tmp_path)
    assert path.exists(), "Path should exist"
    assert path.is_file(), "Path should be a file"

    tiff = TiffImageReader(path)
    assert tiff.dtype == dtype, "Tiff should be float32"
    assert tiff.n_channels == 3, "Tiff should have 3 channels"


def test_write_float_array_as_uint8(tmp_path):
    array = np.random.random((3, 1024, 1024))
    reader = ArrayImageReader(tmp_path, array)
    assert reader.n_channels == 3, "Array should have 3 channels"
    assert not reader.is_rgb, "Array should not be rgb"
    writer = OmeTiffWriter(reader)
    path = writer.write("test", tmp_path, as_uint8=True)
    assert path.exists(), "Path should exist"
    assert path.is_file(), "Path should be a file"

    tiff = TiffImageReader(path)
    assert tiff.dtype == np.uint8, "Tiff should be uint8"
    assert tiff.n_channels == 3, "Tiff should have 3 channels"


def test_write_float_array_as_uint8_channel_ids(tmp_path):
    array = np.random.random((3, 1024, 1024))
    reader = ArrayImageReader(tmp_path, array)
    assert reader.n_channels == 3, "Array should have 3 channels"
    assert not reader.is_rgb, "Array should not be rgb"
    writer = OmeTiffWriter(reader)
    path = writer.write("test", tmp_path, as_uint8=True, channel_ids=(0, 2))
    assert path.exists(), "Path should exist"
    assert path.is_file(), "Path should be a file"

    tiff = TiffImageReader(path)
    assert tiff.dtype == np.uint8, "Tiff should be uint8"
    assert tiff.n_channels == 2, "Tiff should have 2 channels"


def test_write_lazy_multichannel(tmp_path):
    array = np.random.random((3, 1024, 1024))
    channel_names = ["C1", "C2", "C3"]
    resolution = 0.5

    wrapper = OmeTiffWrapper()
    with wrapper.write(
        channel_names=channel_names,
        resolution=resolution,
        dtype=array.dtype,
        shape=array.shape,
        name="test",
        output_dir=tmp_path,
    ):
        for i in range(array.shape[0]):
            wrapper.add_channel(i, channel_names[i], array[:, :, i])
    assert wrapper.path.exists(), "Path should exist"

    tiff = TiffImageReader(wrapper.path)
    assert tiff.dtype == array.dtype, "Tiff should be float32"
    assert tiff.n_channels == 3, "Tiff should have 3 channels"
    assert tiff.resolution == resolution, "Tiff should have resolution"
    assert tiff.channel_names == channel_names, "Tiff should have channel names"


def test_write_lazy_rgb(tmp_path):
    array = np.random.randint(0, 255, (1024, 1024, 3), dtype=np.uint8)

    channel_names = ["R", "G", "B"]
    resolution = 0.5

    wrapper = OmeTiffWrapper()
    with wrapper.write(
        channel_names=channel_names,
        resolution=resolution,
        dtype=array.dtype,
        shape=array.shape,
        name="test",
        output_dir=tmp_path,
    ):
        assert wrapper.reader.is_rgb, "Array should be rgb"
        assert wrapper.is_rgb, "Wrapper should be rgb"
        wrapper.add_channel([0, 1, 2], channel_names, array)
    assert wrapper.path.exists(), "Path should exist"

    tiff = TiffImageReader(wrapper.path)
    assert tiff.dtype == array.dtype, "Tiff should be float32"
    assert tiff.n_channels == 3, "Tiff should have 3 channels"
    assert tiff.resolution == resolution, "Tiff should have resolution"
    assert tiff.channel_names == channel_names, "Tiff should have channel names"
    assert tiff.is_rgb, "Tiff should be rgb"


def test_write_lazy_multichannel_as_uint8(tmp_path):
    array = np.random.random((3, 1024, 1024))
    channel_names = ["C0", "C1", "C2"]
    resolution = 0.5

    wrapper = OmeTiffWrapper()
    with wrapper.write(
        channel_names=channel_names,
        resolution=resolution,
        dtype=array.dtype,
        shape=array.shape,
        name="test",
        output_dir=tmp_path,
        as_uint8=True,
    ):
        for i in range(array.shape[0]):
            wrapper.add_channel(i, channel_names[i], array[i])
    assert wrapper.path.exists(), "Path should exist"

    tiff = TiffImageReader(wrapper.path)
    assert tiff.dtype == np.uint8, "Tiff should be uint8"
    assert tiff.n_channels == 3, "Tiff should have 3 channels"
    assert tiff.resolution == resolution, "Tiff should have resolution"
    assert tiff.channel_names == channel_names, "Tiff should have channel names"


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_write_merge_dtype(tmp_path, dtype):
    array = np.random.random((3, 1024, 1024)).astype(dtype)
    reader = ArrayImageReader(tmp_path, array)

    merge = MergeImages([reader, reader], [1, 1])
    writer = MergeOmeTiffWriter(merge)
    path = writer.write("test", ["reader-1", "reader-2"], tmp_path)
    assert path.exists(), "Path should exist"

    tiff = TiffImageReader(path)
    assert tiff.dtype == dtype, "Tiff should be float32"
    assert tiff.n_channels == 6, "Tiff should have 6 channels"


def test_write_merge_as_uint8(tmp_path):
    array = np.random.random((3, 1024, 1024)).astype(np.float32)
    reader = ArrayImageReader(tmp_path, array)

    merge = MergeImages([reader, reader], [1, 1])
    writer = MergeOmeTiffWriter(merge)
    path = writer.write("test", ["reader-1", "reader-2"], tmp_path, as_uint8=True)
    assert path.exists(), "Path should exist"

    tiff = TiffImageReader(path)
    assert tiff.dtype == np.uint8, "Tiff should be uint8"
    assert tiff.n_channels == 6, "Tiff should have 6 channels"


def test_write_merge_as_uint8_channel_ids(tmp_path):
    array = np.random.random((3, 1024, 1024)).astype(np.float32)
    reader1 = ArrayImageReader(tmp_path, array)
    reader1._channel_names = ["C0", "C1", "C2"]

    reader2 = ArrayImageReader(tmp_path, array)
    reader2._channel_names = ["C3", "C4", "C5"]

    merge = MergeImages([reader1, reader2], [1, 1], channel_names=[reader1.channel_names, reader2.channel_names])
    writer = MergeOmeTiffWriter(merge)
    path = writer.write("test", ["reader-1", "reader-2"], tmp_path, as_uint8=True, channel_ids=[[0, 1], [0, 2]])
    assert path.exists(), "Path should exist"

    tiff = TiffImageReader(path)
    assert tiff.dtype == np.uint8, "Tiff should be uint8"
    assert tiff.n_channels == 4, "Tiff should have 4 channels"
    assert "C0" in tiff.channel_names[0], "Channel name should be C0"
    assert "C1" in tiff.channel_names[1], "Channel name should be C1"
    assert "C3" in tiff.channel_names[2], "Channel name should be C3"
    assert "C5" in tiff.channel_names[3], "Channel name should be C5"


def test_write_merge_as_uint8_channel_ids_not_all(tmp_path):
    array = np.random.random((3, 1024, 1024)).astype(np.float32)
    reader1 = ArrayImageReader(tmp_path, array)
    reader1._channel_names = ["C0", "C1", "C2"]

    reader2 = ArrayImageReader(tmp_path, array)
    reader2._channel_names = ["C3", "C4", "C5"]

    merge = MergeImages([reader1, reader2], [1, 1], channel_names=[reader1.channel_names, reader2.channel_names])
    writer = MergeOmeTiffWriter(merge)
    path = writer.write("test", ["reader-1", "reader-2"], tmp_path, as_uint8=True, channel_ids=[[0, 1], None])
    assert path.exists(), "Path should exist"

    tiff = TiffImageReader(path)
    assert tiff.dtype == np.uint8, "Tiff should be uint8"
    assert tiff.n_channels == 5, "Tiff should have 5 channels"
    assert "C0" in tiff.channel_names[0], "Channel name should be C0"
    assert "C1" in tiff.channel_names[1], "Channel name should be C1"
    assert "C3" in tiff.channel_names[2], "Channel name should be C3"
    assert "C4" in tiff.channel_names[3], "Channel name should be C4"
    assert "C5" in tiff.channel_names[4], "Channel name should be C5"


def test_write_merge_as_uint8_channel_ids_list_of_none(tmp_path):
    array = np.random.random((3, 1024, 1024)).astype(np.float32)
    reader1 = ArrayImageReader(tmp_path, array)
    reader1._channel_names = ["C0", "C1", "C2"]

    reader2 = ArrayImageReader(tmp_path, array)
    reader2._channel_names = ["C3", "C4", "C5"]

    merge = MergeImages([reader1, reader2], [1, 1], channel_names=[reader1.channel_names, reader2.channel_names])
    writer = MergeOmeTiffWriter(merge)
    path = writer.write("test", ["reader-1", "reader-2"], tmp_path, as_uint8=True, channel_ids=[None, None])
    assert path.exists(), "Path should exist"

    tiff = TiffImageReader(path)
    assert tiff.dtype == np.uint8, "Tiff should be uint8"
    assert tiff.n_channels == 6, "Tiff should have 6 channels"
